:- module(conflictSolve_comm_super, [conflictResolution_supervisor/4]).

% conflictResolution_supervisor/4  to be invoked by the supervisor
%  see below for the details...
% conflictResolution_client/4  to be invoked by each client/agent (defined elsewhere)

:- use_module('settings.pl').
:- use_module(library(lists)).
:- use_module(m_lindaclient).



% different ways to resolve conflicts (see settings.pl)
:- arbitrationStrategy(_N, SupervisorModuleFile), use_module(SupervisorModuleFile).




% conflictResolution_supervisor(+List, +StepNumber, -Goods, -Bads)
% determines the subset Goods of List consisting of the agents allowed to act
% those in Bads will be inhibited. (Goods=[] is allowed).

%% APPROACH BASED ON COMMUNICATION AMONG AGENTS 
% (hoping they reach an agreement) 

% We notify the conflicting agents about the conflict and the
% list of involved agents, each of them is supposed to answer
% (possibly after coordination, see below) one among:
% 'RETRY-OP', meaning that he asks again for the same action
%    (this happens to answer to a request of another agent)
% 'NO-OP', meaning that he gives up for this time step.
% Any answer different from 'RETRY-OP' is considered as being 'NO-OP'.
% The supervisor waits for all agents' answers.
% If the collected answers still implies inconsistency than it
% iterates the procedure on the remaining active agents only.
% The other possible answer, 'ARBITRATE', is handled by solveByArbitration.
% One of such answer suffices to invoke arbitration (regardless of the
% answers of other agents).
% Arbitration might involve only those agents that answered 'RETRY-OP'
% or 'ARBITRATE' (see the implementation of solveByArbitration).

conflictResolution_supervisor([A], _, [A], []) :- !.  % ok if singleton.
conflictResolution_supervisor([A,B|List], StepNum,Goods,Bads) :-
	slice([A,B|List],TAs),
	findall(T-ForT, setof(Act, member(T-Act, TAs), ForT), T_As_list),
	delegateConflictResolution(T_As_list,[A,B|List],StepNum,Goods,Bads).

slice([], []).
slice([_P-T-(Act-_Cs)|Rest], [T-Act|R]):-
	slice(Rest, R1).

delegateConflictResolution(T_As_list,PTACsList,StepNum,Goods,Bads) :-
	delegateConflictResolutionAux(T_As_list,StepNum,TckOptList), %gathers agents' choices
	( memberchk(_-'ARBITRATE', TckOptList) % if at leas one is ARBITRATE...
		-> (
		    solveByArbitration(PTACsList, TckOptList, Goods, Bads) 
		    %TODO qualcos'altro ????
		   )
		;  ( % no one wants arbitration
		    splitGoodsBads(TckOptList,PTACsList,Goods1,Bads1), %separa chi insiste(Goods) da chi rinuncia(Bads)
		    ( checkSteadyCondition(Bads1,Goods1)  %test fail se nessuno ha rinunciato
		    	-> ( %verifica consistenza, eventuale ricorsione...
				goodsConsistencyCheck(Goods1)
					-> ( Goods=Goods1, Bads=Bads1 ) 
					;  ( conflictResolution_supervisor(Goods1, StepNum,Goods,Bads2),
						append(Bads1,Bads2,Bads) )
	 		   )
		   ;  ( %if no one gave up (e.g., all posted 'RETRY-OP'), then go for arbitration
				solveByArbitration(PTACsList, TckOptList, Goods, Bads)
			   )
		    )
		   )
	).

delegateConflictResolutionAux(T_As_list,StepNum,TckOptList) :-
	notifyAboutExistingConflict(T_As_list,StepNum),
	expectedTermsTer(T_As_list, List, StepNum, proposedOption(StepNum,_Tck,_Opt)),
	getProposedSolution(T_As_list, List, TckOptList, StepNum).

expectedTermsTer([], [], _StepNum, _Template).
expectedTermsTer([Tck|Tckts],[Term|R],StepNum, Template) :-
	copy_term(Template,Term),
	Term =.. [_,StepNum,Tck|_],
	expectedTermsTer(Tckts, R, StepNum, Template).

 
checkSteadyCondition([],_Goods) :- !,
	% we should go for arbitration since no agent gave up...
	%format(" NESSUNO HA RINUNCIATO (rischio di loop):~w~n", [Goods]),
	!,fail.
checkSteadyCondition(_Bads,[]) :- !,
	% all agents gave up...
	%format(" TUTTI HANNO RINUNCIATO :~w~n", [Bads]),
	true.
checkSteadyCondition([_|_],[_|_]).


goodsConsistencyCheck(PTCsList) :-
	mkTriplesBis(PTCsList,FAVs),
	sort(FAVs, Triples),
	keys_and_values(Triples, Fls, _), sort(Fls, Fluents),!,
	findConflicts(Fluents, Triples, []).	 %consistency means Conflicts=[]

mkTriplesBis([],[]).
mkTriplesBis([_-ID-Cs|PTCsList],Triples) :- 
	(setof(F-(ID-V), member([F,V],Cs), FAVs);(FAVs=[])),!,
	mkTriplesBis(PTCsList,FAVss),
	append(FAVs,FAVss,Triples). 


%the Goods are those answering 'RETRY-OP', any other answer is considered to be 'NO-OP':
splitGoodsBads([], _PTCsList, [],[]).
splitGoodsBads([Tck-'RETRY-OP'|TckOptList], PTCsList, [P-Tck-Cs|Goods], Bads) :- !,
	memberchk(P-Tck-Cs, PTCsList),
	splitGoodsBads(TckOptList, PTCsList, Goods, Bads).
splitGoodsBads([Tck-'NO-OP'|TckOptList], PTCsList, Goods, [P-Tck-Cs|Bads]) :- !,
	memberchk(P-Tck-Cs, PTCsList),
	splitGoodsBads(TckOptList, PTCsList, Goods, Bads).
splitGoodsBads([Tck-_SomethingElse|TckOptList], PTCsList, Goods, [P-Tck-Cs|Bads]) :- !,
	memberchk(P-Tck-Cs, PTCsList),
	splitGoodsBads(TckOptList, PTCsList, Goods, Bads).


getProposedSolution([], _List, [], _).   DA AGGIORNARE/VERIFICARE  -- SE SERVE PER IL calcio???
getProposedSolution(Tckts, List, [ATck-AOpt|Opts], StepNum) :-
	m_in(List, proposedOption(StepNum,ATck,AOpt)),
	%format("        RICEVUTA PROPOSED SOLUTION :~w~n", [proposedOption(StepNum,ATck,AOpt)]),
	selectchk(ATck, Tckts, TcktsRest),
	getProposedSolution(TcktsRest, List, Opts, StepNum).

notifyAboutExistingConflict(T_As_list,StepNum) :-
	notifyAboutExistingConflictAux(T_As_list,T_As_list,StepNum).
notifyAboutExistingConflictAux([],_,_).
notifyAboutExistingConflictAux([Tck-Acts|T_As_list],AllT_As_list,StepNum) :-
	m_out(actionResponse(StepNum,Tck,call_for_agreement(Acts-AllT_As_list))),
	notifyAboutExistingConflictAux(T_As_list,AllT_As_list,StepNum).


%Res is a list of  F-ConflSubset where F is a Fluent and ConflSubset is the list of Agents disagreeing on F
findConflicts([], _, []).
findConflicts([F|Fs], Triples, Res) :-
	findall(A-V, member(F-(A-V),Triples), AVsForF),
	(isConflicting(AVsForF, ConflSubset) -> (Res=[F-ConflSubset|Res1])
					     ;  (Res=Res1) ),
	findConflicts(Fs, Triples, Res1).



isConflicting(AVsForF, ConflSubset) :-
	keys_and_values(AVsForF, As, Vs),!,
	(sort(Vs,[_]) -> fail ; sort(As,ConflSubset)).



